import logging

from .models import Endpoint
from . import client

log = logging.getLogger(__name__)


def configure(
    name,
    settings,
    prefix="zipkin.",
    use_requests=True,
    use_celery=True,
    use_xmlrpclib=True,
):
    """Include the zipkin definitions"""

    endpoint = Endpoint(name)
    client.configure(settings, prefix=prefix)

    # Install in libs here
    if use_requests:
        from .binding.requests import bind as bind_requests

        bind_requests()

    if use_celery:
        from .binding.celery import bind as bind_celery

        bind_celery(endpoint)

    if use_xmlrpclib:
        from .binding.xmlrpclib import bind as bind_xmlrpclib

        bind_xmlrpclib()

    return endpoint



# export the API here
from .api import trace, get_current_trace, stack_trace, Trace

from .config import configure
from .thread import local  # XXX remove me from here


__version__ = "0.8.2"

__all__ = ["trace", "Trace", "get_current_trace", "stack_trace", "configure", "local"]



try:
    from .impl import bind
except ImportError as exc:
    import logging

    logging.getLogger(__name__).warn("SQLAlchemy not installed")

    def bind(*args, **kwargs):
        pass

    raise



import logging

from celery.signals import before_task_publish, task_prerun, task_postrun

from zipkin.models import Endpoint
from . import events

log = logging.getLogger(__name__)


def bind(endpoint=None):
    if not endpoint:
        endpoint = Endpoint("Celery")

    events.endpoint = endpoint

    log.info("Attaching zipkin to celery signals")
    before_task_publish.connect(events.task_send_handler)
    task_prerun.connect(events.task_prerun_handler)
    task_postrun.connect(events.task_postrun_handler)
    log.info("zipkin signals attached")



import logging

from flask import request_started, request_finished

from zipkin.models import Endpoint
from . import events

log = logging.getLogger(__name__)


def bind(app, endpoint=None):
    if not endpoint:
        endpoint = Endpoint(app.name)

    events.endpoint = endpoint

    log.info("Attaching zipkin to Flask signals")
    request_started.connect(events.pre_request, app)
    request_finished.connect(events.pre_response, app)
    log.info("zipkin signals attached")



# flake8: noqa
try:
    from requests.packages.urllib3.packages.six.moves import range

except ImportError:
    """Remove embedded version of six from request.packages v0.14...
    Missing six.moves.range
    """

    import sys
    import os.path
    import inspect

    try:
        import requests.packages

        if "six.moves" in sys.modules:
            del sys.modules["six.moves"]
            dirname = os.path.dirname(inspect.getfile(requests.packages))
            if dirname in sys.path:
                sys.path.remove(dirname)
    except ImportError:
        pass

    import six.moves



import pkg_resources
import thriftpy2

filename = pkg_resources.resource_filename("zipkin", "resources/zipkinCore.thrift")
zipkincore_thrift = thriftpy2.load(filename, module_name="zipkincore_thrift")



import unittest

from six import u

from ..models import TraceStack, Annotation, Trace


class TestTrace(unittest.TestCase):
    def setUp(self):
        self.trace = Trace("test")

    def test_child(self):
        self.trace.child("child")

        self.assertEqual(len(self.trace.children()), 2)
        self.assertEqual(self.trace.children()[0].name, "child")

    def test_child_noref(self):
        self.trace.child_noref("test")
        self.assertEqual(len(self.trace.children()), 1)

    def test_record(self):
        self.trace.record(Annotation.string("key", "value"))
        self.assertEqual(self.trace.children()[0].annotations[0].name, "key")

    def test_type(self):
        # Unicode or string should not raise
        self.trace.child("string")
        self.trace.child(u("unicode"))

        # rest should raise
        self.assertRaises(AssertionError, self.trace.child, 1)
        self.assertRaises(AssertionError, self.trace.child, None)


class TestTraceStack(unittest.TestCase):
    def setUp(self):
        self.stack = TraceStack()

    def test_push_pop(self):
        self.stack.append(Trace("testing stack"))
        self.stack.child("subtrace")

        self.assertEqual(self.stack.current.name, "subtrace")
        trace = self.stack.pop()
        self.assertEqual(trace.name, "subtrace")

        self.assertEqual(self.stack.current.name, "testing stack")

    def test_pop_empty(self):
        self.assertRaises(IndexError, self.stack.pop)

    def test_reset(self):
        self.stack.append(Trace("testing stack"))
        self.stack.reset()
        self.assertEqual(len(self.stack.stack), 0)
        self.assertEqual(self.stack.current, None)

    def test_replace(self):
        self.stack.append(Trace("testing stack"))
        self.stack.replace(Trace("testing stack"))

        # We replaced stack, we should have only one item
        self.assertEqual(len(self.stack.stack), 1)

    def test_types(self):
        self.assertRaises(AssertionError, self.stack.replace, "string")
        self.assertRaises(AssertionError, self.stack.append, "string")
        self.stack.append(Trace("testing stack"))
        self.assertRaises(AssertionError, self.stack.child, 1)
        self.assertRaises(AssertionError, self.stack.child, None)
        self.assertRaises(AssertionError, self.stack.child, "string", "test")
        self.assertRaises(AssertionError, self.stack.child, "string", False)

        # String or unicode should not raise
        self.stack.child("string")
        self.stack.child(u("unicode"))



import requests.sessions

from . import events
from .events import request_adapter


def bind():
    if not getattr(requests.sessions.Session, "_zipkin_patched", False):
        old_init = requests.sessions.Session.__init__
        requests.sessions.Session.__init__ = events.session_init(old_init)
        requests.sessions.Session._zipkin_patched = True



try:
    from .impl import bind, request_adapter
except ImportError as exc:
    import logging

    logging.getLogger(__name__).warn("requests not installed")

    def bind():
        pass

    def request_adapter(adapter):
        return adapter



"""
Django App to trace call with zipkin.


Usage:

You need to add this in your django settings.

::

    ZIPKIN_SERVICE_NAME = "my-service-name"
    ZIPKIN_COLLECTOR = "zipkin.localdomain"  # hostname of the zipkin server
    ZIPKIN_ENV = "dev"  # Environment
    ZIPKIN_BINDINGS = ["requests"]  # List of library to patch/attach

    INSTALLED_APPS = [
        ...,
        "zipkin.binding.django",
        ...
    ]


"""

from django.conf import settings
from django.apps import AppConfig
from zipkin.config import configure


def get_settings():
    zk_settings = {
        "zipkin.transport": getattr(settings, "ZIPKIN_TRANSPORT", "scribe"),
        "zipkin.collector": settings.ZIPKIN_COLLECTOR,
        "zipkin.service_name": settings.ZIPKIN_SERVICE_NAME,
        "zipkin.env": settings.ZIPKIN_ENV,
    }
    if hasattr(settings, "ZIPKIN_COLLECTOR_PORT"):
        zk_settings["zipkin.collector.port"] = settings.ZIPKIN_COLLECTOR_PORT
    if hasattr(settings, "ZIPKIN_COLLECTOR_SCHEME"):
        zk_settings["zipkin.collector.scheme"] = settings.ZIPKIN_COLLECTOR_SCHEME
    if hasattr(settings, "ZIPKIN_TRANSPORT_ASYNC"):
        zk_settings["zipkin.transport.async"] = settings.ZIPKIN_TRANSPORT_ASYNC

    return zk_settings


class ZipkinConfig(AppConfig):
    """
    Configure zipkin from django settings.
    """

    name = "zipkin"
    service_name = ""

    def ready(self):

        bindings = []  # settings.ZIPKIN_BINDINGS
        appsettings = get_settings()
        ZipkinConfig.service_name = "%s-%s" % (
            settings.ZIPKIN_SERVICE_NAME,
            settings.ZIPKIN_ENV,
        )
        configure(
            appsettings["zipkin.service_name"],
            appsettings,
            use_celery="celery" in bindings,
            use_requests="request" in bindings,
            use_xmlrpclib="xmlrpclib" in bindings,
        )



import unittest
import threading
import socket
import select

from ..transport.scribeclient import Client
from ..models import Trace, Annotation


class Sinkhole(threading.Thread):
    def __init__(self, *args, **kwargs):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.bind(("127.0.0.1", 0))
        self.port = self.sock.getsockname()[1]
        self.sock.listen(5)

        self.stopped = False

        super(Sinkhole, self).__init__()

        self.has_received_data = False
        self.has_data = threading.Lock()
        self.has_data.acquire()

    def run(self):
        sockets = [self.sock]

        while not self.stopped:
            readable, _, _ = select.select(sockets, [], [])

            for s in readable:
                if s is self.sock:
                    connection, _ = s.accept()
                    connection.setblocking(0)
                    sockets.append(connection)
                else:
                    data = s.recv(1024)
                    if not data:
                        sockets.remove(s)
                        s.close()
                    else:
                        self.has_received_data = True
                        self.has_data.release()

        self.sock.close()

    def stop(self):
        self.stopped = True
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect_ex(("127.0.0.1", self.port))
        s.close()
        self.join()


class TestClient(unittest.TestCase):
    def setUp(self):
        self.server = Sinkhole()
        self.server.start()

    def tearDown(self):
        self.server.stop()

    def test_simple(self):
        Client.configure(
            {
                "collector": "127.0.0.1",
                "collector.port": self.server.port,
            },
            "",
        )

        trace = Trace("test")
        trace.record(Annotation.string("foo", "bar"))
        Client.get_connection()

        import time

        time.sleep(1)  # Just to give it time to connect

        Client.log(trace)
        self.server.has_data.acquire()
        self.server.has_data.release()
        self.assertTrue(self.server.has_received_data)
        Client.disconnect()



from functools import wraps

from .models import Annotation
from .thread import local

__all__ = ["trace", "get_current_trace"]


class Trace(object):
    def __init__(self, name):
        self.name = name
        self.recording = None
        self.trace = None

    def __enter__(self):

        try:
            self.trace = local().child(self.name)
            annotation = Annotation.server_recv()
            self.trace.record(annotation)
            self.recording = True
        except Exception:
            self.recording = False

    def __exit__(self, type, value, traceback):
        if self.recording:
            self.trace.record(Annotation.server_send())
            local().pop()

    def __call__(self, func):
        @wraps(func)
        def decorated(*args, **kwds):
            with self:
                return func(*args, **kwds)

        return decorated


def trace(name):
    """A decorator that trace the decorated function"""

    if hasattr(name, "__call__"):
        return trace(name.__name__)(name)

    def func_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with Trace(name):
                return func(*args, **kwargs)

        return wrapper

    return func_decorator


def get_current_trace():
    return local().current


def stack_trace(trace):
    return local().append(trace)



"""Middleware for django."""
import time

from django.conf import settings

from zipkin import local
from zipkin.api import get_current_trace, stack_trace
from zipkin.models import Trace, Annotation, Endpoint
from zipkin.util import int_or_none
from zipkin.client import log as zipkin_log

from .apps import ZipkinConfig


def init_trace(request):
    headers = request.headers
    trace_name = request.method + " " + request.path_info
    trace = Trace(
        trace_name,
        int_or_none(headers.get("X-B3-TraceId", None)),
        int_or_none(headers.get("X-B3-SpanId", None)),
        int_or_none(headers.get("X-B3-ParentSpanId", None)),
        endpoint=Endpoint(ZipkinConfig.service_name),
    )
    trace.record(Annotation.string("http.path", request.path_info))
    trace.record(Annotation.string("span.kind", "client"))
    trace.record(Annotation.server_recv())
    stack_trace(trace)
    return trace


def log_response(trace, response):
    trace.record(
        Annotation.string("http.responsecode", "{0}".format(response.status_code))
    )
    trace.record(Annotation.server_send())
    zipkin_log(trace)
    local().reset()


def add_header_response(response):
    response["Trace-Id"] = str(get_current_trace().trace_id)


def zk_middleware(get_response):
    """
    Zipkin Middleware to add in the django settings.


    Usage:

    ::

        MIDDLEWARE = [
            "zipkin.binding.django.middleware.zk_middleware",
            ...
        ]
    """

    def middleware(request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        trace = init_trace(request)

        response = get_response(request)

        add_header_response(response)
        log_response(trace, response)

        return response

    return middleware


def zk_slow_trace_middleware(get_response):
    """
    Zipkin Middleware to trace slow query only added in the django settings.

    Usage:

    ::
        # Only send trace of query that take more than 1.5 seconds
        ZIPKIN_SLOW_LOG_DURATION_EXCEED = 1.5

        MIDDLEWARE = [
            "zipkin.binding.django.middleware.zk_slow_trace_middleware",
            ...
        ]
    """

    def middleware(request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        start = time.time()
        trace = init_trace(request)

        response = get_response(request)
        duration = time.time() - start
        if duration >= settings.ZIPKIN_SLOW_LOG_DURATION_EXCEED:
            add_header_response(response)
            log_response(trace, response)
        return response

    return middleware



from __future__ import absolute_import

import re
import six.moves.xmlrpc_client as xmlrpclib
import logging

from zipkin import local
from zipkin.models import Annotation, Endpoint
from zipkin.util import hex_str


log = logging.getLogger(__name__)

_parse_method_name = re.compile(r"<methodName>([^<]*)</methodName>", re.I)


class MonkeyTransport(xmlrpclib.Transport):
    """Monkey patched version of xmlrpclib.Transport to plug zipkin in"""

    __origin = xmlrpclib.Transport

    def request(self, host, handler, request_body, verbose=0):
        try:
            https = isinstance(self, xmlrpclib.SafeTransport)
            protocol = "https://" if https else "http://"
            target = "%s%s%s" % (protocol, host, handler)

            match = _parse_method_name.search(request_body)
            method = match.group(1) if match else None

            parent_trace = local().current
            self._trace = parent_trace.child("xmlrpclib")

            self._trace.record(Annotation.string("uri", target))
            if method:
                self._trace.record(Annotation.string("method", method))
            self._trace.record(Annotation.server_recv())
        except Exception as exc:
            log.error(repr(exc))

        try:
            return self.__origin.request(self, host, handler, request_body, verbose)
        finally:
            try:
                self._trace.record(Annotation.server_send())
            except:
                pass

    def send_host(self, connection, host):
        ret = self.__origin.send_host(self, connection, host)
        try:
            forward = self.trace.child_noref("subservice")
            connection.putheader("X-B3-TraceId", hex_str(forward.trace_id))
            connection.putheader("X-B3-SpanId", hex_str(forward.span_id))
            if forward.parent_span_id is not None:
                connection.putheader(
                    "X-B3-ParentSpanId", hex_str(forward.parent_span_id)
                )
        finally:
            return ret


def bind(endpoint=None):
    log.info("Binding zipkin to xmlrpclib")
    if not endpoint:
        endpoint = Endpoint("xmlrpc")

    xmlrpclib.Transport = MonkeyTransport

    log.info("zipkin bound to xmlrpclib")


def unbind():
    xmlrpclib.Transport = MonkeyTransport.__origin



from __future__ import unicode_literals, absolute_import

import os
import unittest
from mock import patch

from flask import Flask
from flask.views import View
import requests

import zipkin
from ...binding.flask import bind as bind_zipkin
from ...binding.requests import bind as bind_requests
from ...models import Endpoint, Trace
from ...thread import local

from ..utils import DummyClient, dummy_log, Server


class Home(View):
    def dispatch_request(self):
        return "Guten Morgen!"


class Forbidden(View):
    def dispatch_request(self):
        return "Verboten!", 403


class FlaskApp(Flask):
    def __init__(self, *args, **kwargs):
        super(FlaskApp, self).__init__(__name__)
        self.setup_url()
        self.setup_zipkin()

    def setup_zipkin(self):
        endpoint = zipkin.configure(
            "My test flask application",
            {"zipkin.collector": "localhost", "zipkin.service_name": "my app"},
        )
        bind_zipkin(self, endpoint)

    def setup_url(self):
        self.add_url_rule("/", methods=["GET"], view_func=Home.as_view(str("home")))
        self.add_url_rule(
            "/interdit", methods=["GET"], view_func=Forbidden.as_view(str("proibito"))
        )


class FlaskAppTestCase(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(FlaskAppTestCase, self).__init__(methodName)
        self.logger = patch("zipkin.binding.flask.events.log", dummy_log)

    def setUp(self):
        self.logger.start()
        DummyClient.reset()

        self.app = FlaskApp()
        self.client = self.app.test_client()

    def tearDown(self):
        self.logger.stop()

    def test_home(self):
        ret = self.client.get("/")
        self.assertEqual(ret.data, b"Guten Morgen!")

        traces = DummyClient._client.messages
        self.assertEqual(
            len(traces),
            1,
            "There should be one trace for " "the request just processed",
        )


class FlaskAppRealServerTestCase(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(FlaskAppRealServerTestCase, self).__init__(methodName)
        self.logger = patch("zipkin.binding.flask.events.log", dummy_log)
        bind_requests()

    def setUp(self):
        self.http_proxy = os.environ.pop("http_proxy", None)
        self.https_proxy = os.environ.pop("https_proxy", None)

        self.logger.start()
        DummyClient.reset()

        self.app = Server(FlaskApp())
        self.app.start()

    def tearDown(self):
        self.logger.stop()
        self.app.stop()

        if self.http_proxy:
            os.environ["http_proxy"] = self.http_proxy
        if self.https_proxy:
            os.environ["https_proxy"] = self.https_proxy

    def test_home_no_headers(self):
        ret = requests.get(self.app.url)
        self.assertEqual(ret.text, "Guten Morgen!")

        traces = DummyClient._client.messages
        self.assertEqual(
            len(traces),
            1,
            "There should be one trace for " "the request just processed",
        )

    def test_home_headers(self):
        endpoint = Endpoint("127.0.0.1", 0, "RequestsWithContextTestCase")
        trace = Trace("flask", None, None, None, endpoint=endpoint)
        local().replace(trace)

        ret = requests.get(self.app.url)
        self.assertEqual(ret.text, "Guten Morgen!")

        traces = DummyClient._client.messages
        self.assertEqual(
            len(traces),
            1,
            "There should be one trace for " "the request just processed",
        )
        self.assertEqual(
            traces[0].trace_id,
            trace.trace_id,
            "TraceId should be read from client if present",
        )



from __future__ import unicode_literals

import threading

import json
from wsgiref.simple_server import make_server, WSGIRequestHandler
from wsgiref.handlers import SimpleHandler
from six.moves.urllib.parse import urljoin
from flask import Flask, request
from flask.views import View


class Headers(View):
    def dispatch_request(self):
        return json.dumps({"headers": dict(request.headers)})


class FlaskApp(Flask):
    def __init__(self, *args, **kwargs):
        super(FlaskApp, self).__init__(__name__)
        self.setup_url()

    def setup_url(self):
        self.add_url_rule(
            "/headers", methods=["GET"], view_func=Headers.as_view(str("headers"))
        )


httpbin_app = FlaskApp()


class ScribeClient(object):
    def __init__(self):
        self.messages = []

    def Log(self, messages):
        self.messages.append(messages)


class DummyClient(object):
    _client = ScribeClient()

    @classmethod
    def configure(cls, settings, prefix):
        pass

    @classmethod
    def get_connection(cls):
        return cls._client

    @classmethod
    def log(cls, trace):
        return cls._client.Log(trace)

    @classmethod
    def reset(cls):
        cls._client = ScribeClient()


def dummy_log(trace):
    DummyClient.log(trace)


class ServerHandler(SimpleHandler):

    server_software = "HTTPBIN/0.1.0"
    http_version = "1.1"

    def cleanup_headers(self):
        SimpleHandler.cleanup_headers(self)
        self.headers["Connection"] = "Close"

    def close(self):
        try:
            self.request_handler.log_request(
                self.status.split(" ", 1)[0], self.bytes_sent
            )
        finally:
            SimpleHandler.close(self)


class Handler(WSGIRequestHandler):
    def handle(self):
        """Handle a single HTTP request"""

        self.raw_requestline = self.rfile.readline()
        if not self.parse_request():  # An error code has been sent, just exit
            return

        handler = ServerHandler(
            self.rfile, self.wfile, self.get_stderr(), self.get_environ()
        )
        handler.request_handler = self  # backpointer for logging
        handler.run(self.server.get_app())

    def get_environ(self):
        """
        wsgiref simple server adds content-type text/plain to everything, this
        removes it if it's not actually in the headers.
        """
        # Note: Can't use super since this is an oldstyle class in python 2.x
        environ = WSGIRequestHandler.get_environ(self).copy()
        if self.headers.get("content-type") is None:
            del environ["CONTENT_TYPE"]
        return environ


class Server(threading.Thread):
    """
    HTTP server running a WSGI application in its own thread.
    """

    def __init__(self, application=httpbin_app, host="127.0.0.1", port=0, **kwargs):
        self.app = application
        self._server = make_server(
            host, port, self.app, handler_class=Handler, **kwargs
        )
        self.host = self._server.server_address[0]

        self.port = self._server.server_address[1]
        self.protocol = "http"

        super(Server, self).__init__(
            name=self.__class__,
            target=self._server.serve_forever,
        )

    def __del__(self):
        self.stop()

    def __add__(self, other):
        return self.url + other

    def stop(self):
        self._server.shutdown()

    @property
    def url(self):
        return "{0}://{1}:{2}".format(self.protocol, self.host, self.port)

    def join(self, url, allow_fragments=True):
        return urljoin(self.url, url, allow_fragments=allow_fragments)



import logging

import requests
from requests.exceptions import Timeout, ConnectionError, ConnectTimeout

from ..util import base64_thrift_formatter_many


logger = logging.getLogger(__name__)


class Client(object):

    host = None
    port = 9411
    scheme = "http"
    _url = None
    _socket_timeout = 1000

    @classmethod
    def configure(cls, settings, prefix):
        cls.host = settings.get(prefix + "collector")
        if prefix + "collector.port" in settings:
            cls.port = int(settings[prefix + "collector.port"])
        if prefix + "collector.scheme" in settings:
            cls.scheme = settings[prefix + "collector.scheme"]
        if prefix + "transport.socket_timeout" in settings:
            cls._socket_timeout = int(settings[prefix + "transport.socket_timeout"])

        cls._url = "%s://%s:%s/api/v1/spans" % (cls.scheme, cls.host, cls.port)

    @classmethod
    def log(cls, trace):
        if cls._url is None:
            raise ValueError("Unconfigured client")
        payload = base64_thrift_formatter_many(trace)
        try:
            requests.post(
                cls._url,
                headers={"Content-Type": "application/x-thrift"},
                data=payload,
                timeout=cls._socket_timeout,
            )
        except ConnectTimeout:
            logger.error("Connect timeout while connecting to zipking collector")
        except ConnectionError:
            logger.error("Cannot connect to zipking collector")
        except Timeout:
            logger.error("Timeout while posting trace")



import random
import struct
import socket

from base64 import b64encode
from six import text_type
from thriftpy2.protocol import TBinaryProtocol
from thriftpy2.thrift import TType
from thriftpy2.protocol.binary import write_list_begin
from thriftpy2.transport import TMemoryBuffer
from thriftpy2.thrift import TDecodeException

from .zipkin import zipkincore_thrift as ttypes


def int_or_none(val):
    if val is None:
        return None

    return int(val, 16)


def hex_str(n):
    return "%0.16x" % (n,)


def uniq_id():
    """
    Create a random 64-bit signed integer appropriate
    for use as trace and span IDs.

    @returns C{int}
    """
    return random.randint(0, (2 ** 64) - 1)


def base64_thrift(thrift_obj):
    trans = TMemoryBuffer()
    tbp = TBinaryProtocol(trans)

    thrift_obj.write(tbp)

    return b64encode(bytes(trans.getvalue())).strip()


def ipv4_to_int(ipv4):
    return struct.unpack("!i", socket.inet_aton(ipv4))[0]


def binary_annotation_formatter(annotation, host=None):
    annotation_types = {
        "string": ttypes.AnnotationType.STRING,
        "bytes": ttypes.AnnotationType.BYTES,
    }

    annotation_type = annotation_types[annotation.annotation_type]

    value = annotation.value

    if isinstance(value, text_type):
        value = value.encode("utf-8")

    return ttypes.BinaryAnnotation(annotation.name, value, annotation_type, host)


def base64_thrift_formatter(trace, annotations):
    thrift_annotations = []
    binary_annotations = []

    try:
        for annotation in annotations:
            host = None
            if annotation.endpoint:
                host = ttypes.Endpoint(
                    ipv4=ipv4_to_int(annotation.endpoint.ip),
                    port=annotation.endpoint.port,
                    service_name=annotation.endpoint.service_name,
                )

            if annotation.annotation_type == "timestamp":
                thrift_annotations.append(
                    ttypes.Annotation(
                        timestamp=annotation.value, value=annotation.name, host=host
                    )
                )
            else:
                binary_annotations.append(binary_annotation_formatter(annotation, host))

        thrift_trace = ttypes.Span(
            name=trace.name,
            trace_id=u64_as_i64(trace.trace_id),
            id=u64_as_i64(trace.span_id),
            parent_id=u64_as_i64(trace.parent_span_id),
            annotations=thrift_annotations,
            binary_annotations=binary_annotations,
        )

        return base64_thrift(thrift_trace)
    except TDecodeException as e:
        raise ValueError(e)


def u64_as_i64(value):
    if not value:
        return value

    try:
        data = struct.pack(">Q", value)
        data = struct.unpack(">q", data)
        return data[0]
    except struct.error as e:
        raise ValueError(e)


def span_to_bytes(thrift_span):
    """
    Returns a TBinaryProtocol encoded Thrift span.

    :param thrift_span: thrift object to encode.
    :returns: thrift object in TBinaryProtocol format bytes.
    """
    transport = TMemoryBuffer()
    protocol = TBinaryProtocol(transport)
    thrift_span.write(protocol)

    return bytes(transport.getvalue())


def base64_thrift_formatter_many(parent_trace):
    """
    Returns a TBinaryProtocol encoded list of Thrift objects.
    :param binary_thrift_obj_list: list of TBinaryProtocol objects to encode.
    :returns: bynary object representing the encoded list.
    """
    traces = list(parent_trace.children())
    transport = TMemoryBuffer()
    write_list_begin(transport, TType.STRUCT, len(traces))
    for trace in traces:
        thrift_annotations = []
        binary_annotations = []

        for annotation in trace.annotations:
            host = None
            if annotation.endpoint:
                host = ttypes.Endpoint(
                    ipv4=ipv4_to_int(annotation.endpoint.ip),
                    port=annotation.endpoint.port,
                    service_name=annotation.endpoint.service_name,
                )

            if annotation.annotation_type == "timestamp":
                thrift_annotations.append(
                    ttypes.Annotation(
                        timestamp=annotation.value, value=annotation.name, host=host
                    )
                )
            else:
                binary_annotations.append(binary_annotation_formatter(annotation, host))

        thrift_trace = ttypes.Span(
            name=trace.name,
            trace_id=u64_as_i64(trace.trace_id),
            id=u64_as_i64(trace.span_id),
            parent_id=u64_as_i64(trace.parent_span_id),
            annotations=thrift_annotations,
            binary_annotations=binary_annotations,
        )

        transport.write(span_to_bytes(thrift_trace))

    return bytes(transport.getvalue())



import errno
import logging
import socket
import struct
from io import BytesIO

from thriftpy2.protocol import TBinaryProtocolFactory
from thriftpy2.protocol.binary import write_message_begin, write_val
from thriftpy2.thrift import TClient, TMessageType, TType
from thriftpy2.transport import TFramedTransportFactory, TSocket, TTransportException

from ..client import Local
from ..util import base64_thrift_formatter
from .scribe import scribe_thrift

logger = logging.getLogger(__name__)

CONNECTION_RETRIES = [1, 10, 20, 50, 100, 200, 400, 1000]

try:
    MSG_NOSIGNAL = socket.MSG_NOSIGNAL
except:
    MSG_NOSIGNAL = 16384  # python2


class TNonBlockingSocket(TSocket):
    def _init_sock(self):
        super(TNonBlockingSocket, self)._init_sock()

        # 1M sendq buffer
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1024 * 1024)

        self.sock.setblocking(False)

    def open(self):
        self._init_sock()

        addr = self.unix_socket or (self.host, self.port)
        status = self.sock.connect_ex(addr)
        try:
            Local.local_ip = self.sock.getsockname()[0]
        except Exception:
            pass

        if status not in [errno.EINPROGRESS, errno.EALREADY]:
            raise IOError(
                "connection attempt on a non-clean socket", errno.errorcode[status]
            )

    def write(self, buff):
        # First ensure our incoming end is always empty
        self.read_all()

        # Then actually try to write to socket
        try:
            # We are a library, we can't just set sighandlers. But we don't
            # want SIGPIPE if peer has gone away either. We better set
            # MSG_NOSIGNAL to avoid that.
            # If peer has disconnected, then a errno.EPIPE will raise, and
            # will be catched on the uppper layer

            self.sock.sendall(buff, MSG_NOSIGNAL)
        except socket.error as e:
            if e.errno not in [
                errno.EINPROGRESS,  # Not connected yet
                errno.EWOULDBLOCK,
            ]:  # write buffer full
                # In all other cases, raise.
                raise

            # If not yet connected or write buffer is full, silently drop.

    def read_all(self):
        """
        Flush incoming buffer
        """
        try:
            receiving = " "
            while len(receiving) > 0:  # socket.error.errno.EAGAIN will exit this
                receiving = self.sock.recv(1024)
        except socket.error as e:
            # if EAGAIN or EWOULDBLOCK, then there is nothing to read
            if e.errno not in [errno.EAGAIN, errno.EWOULDBLOCK]:
                # Otherwise that's an error.
                raise

            return  # No more data to read, or connection is not ready

    def read(self, _):
        """
        Mock response, we don't care about results. We never actually read
        them. But we don't want client to wait for server to reply.
        """
        buffer = BytesIO()
        seq_id = 0  # Sequence id is never compared to message.
        write_message_begin(buffer, "Log_result", TMessageType.REPLY, seq_id)

        response = scribe_thrift.Scribe.Log_result(success=scribe_thrift.ResultCode.OK)
        write_val(buffer, TType.STRUCT, response)

        out = buffer.getvalue()

        # Framed message, starts with length of message.
        return struct.pack("!i", len(out)) + out


def make_client(
    service,
    host,
    port,
    proto_factory=TBinaryProtocolFactory(),
    trans_factory=TFramedTransportFactory(),
    socket_factory=TNonBlockingSocket,
    socket_timeout=1000,
):

    socket = socket_factory(host, port, socket_timeout=socket_timeout)
    transport = trans_factory.get_transport(socket)
    protocol = proto_factory.get_protocol(transport)
    transport.open()
    return TClient(service, protocol)


class Client(object):

    host = None
    port = 9410
    _client = None
    _connection_attempts = 0
    _socket_factory = TNonBlockingSocket
    _socket_timeout = 1000

    @classmethod
    def configure(cls, settings, prefix):
        cls.host = settings.get(prefix + "collector")
        if prefix + "collector.port" in settings:
            cls.port = int(settings[prefix + "collector.port"])
        if prefix + "transport.async" in settings:
            if settings[prefix + "transport.async"].lower() == "false":
                cls._socket_factory = TSocket

        if prefix + "transport.socket_timeout" in settings:
            cls._socket_timeout = int(settings[prefix + "transport.socket_timeout"])

    @classmethod
    def get_connection(cls):
        if not cls._client:
            cls._connection_attempts += 1

            max_retries = CONNECTION_RETRIES[-1]
            if (cls._connection_attempts > max_retries) and not (
                (cls._connection_attempts % max_retries) == 0
            ):
                return
            if (cls._connection_attempts < max_retries) and (
                cls._connection_attempts not in CONNECTION_RETRIES
            ):
                return

            try:
                cls._client = make_client(
                    scribe_thrift.Scribe,
                    host=cls.host,
                    port=cls.port,
                    socket_factory=cls._socket_factory,
                    socket_timeout=cls._socket_timeout,
                )

                cls._connection_attempts = 0
            except TTransportException:
                cls._client = None
                logger.error(
                    "Can't connect to zipkin collector %s:%d" % (cls.host, cls.port)
                )
            except Exception:
                cls._client = None
                logger.exception(
                    "Can't connect to zipkin collector %s:%d" % (cls.host, cls.port)
                )
        return cls._client

    @classmethod
    def log(cls, trace):
        if not cls.host:
            logger.debug("Zipkin tracing is disabled")
            return

        logger.info("logging trace %s", trace.trace_id)

        unknown = (
            "Unknown Exception while logging a trace on "
            "zipkin collector %s:%d" % (cls.host, cls.port)
        )

        client = cls.get_connection()
        if client:
            try:
                messages = [
                    base64_thrift_formatter(t, t.annotations) for t in trace.children()
                ]
                log_entries = [
                    scribe_thrift.LogEntry("zipkin", message) for message in messages
                ]
            except ValueError:
                logger.exception("Error while serializing trace")
                return

            try:
                client.Log(messages=log_entries)
            except EOFError:
                cls._client = None
                logger.error(
                    "EOFError while logging a trace on zipkin "
                    "collector %s:%d" % (cls.host, cls.port)
                )
            except socket.error as err:
                cls._client = None
                if err.errno == errno.EPIPE:
                    logger.error(
                        "Broken pipe while logging a trace "
                        "on zipkin collector %s:%d",
                        cls.host,
                        cls.port,
                    )
                else:
                    logger.exception(unknown)
            except Exception:
                cls._client = None
                logger.exception(unknown)
        else:
            logger.warning("Can't log zipkin trace, not connected")

    @classmethod
    def disconnect(cls):
        if cls._client:
            cls._client.close()



from __future__ import absolute_import

import unittest
import os

import requests

from ..utils import Server
from ...binding.requests import bind as bind_requests
from ...models import Endpoint, Trace
from ...thread import local
from ...util import hex_str


def request_json(response):
    """Signature for response.json varies between
    requests versions, this wrapper finds out which api
    to use
    """
    if hasattr(response.json, "__call__"):
        return response.json()
    else:
        return response.json


class RequestsTestCase(unittest.TestCase):
    """
    This test class starts up a HttpBin instance locally for test-purpose
    """

    def __init__(self, methodName="runTest"):
        super(RequestsTestCase, self).__init__(methodName)
        bind_requests()

    def setUp(self):
        self.http_proxy = os.environ.pop("http_proxy", None)
        self.https_proxy = os.environ.pop("https_proxy", None)

        self.httpbin = Server()
        self.httpbin.start()

    def tearDown(self):
        self.httpbin.stop()
        if self.http_proxy:
            os.environ["http_proxy"] = self.http_proxy
        if self.https_proxy:
            os.environ["https_proxy"] = self.https_proxy


class RequestsNoContextTestCase(RequestsTestCase):
    def test_http_headers_no_context(self):
        resp = requests.get(self.httpbin.url + "/headers")
        headers = request_json(resp)

        self.assertEqual(
            headers.get("X-B3-TraceId", None),
            None,
            "There should be no B3 headers if context " "is not present",
        )


class RequestsWithContextTestCase(RequestsTestCase):
    def setUp(self):
        super(RequestsWithContextTestCase, self).setUp()

        endpoint = Endpoint("127.0.0.1", 0, "RequestsWithContextTestCase")
        self.trace = Trace("requests", None, None, None, endpoint=endpoint)
        local().replace(self.trace)

    def tearDown(self):
        super(RequestsWithContextTestCase, self).tearDown()
        local().reset()

    def test_http_headers_context(self):
        resp = requests.get(self.httpbin.url + "/headers")
        headers = request_json(resp)["headers"]

        self.assertEqual(
            headers.get("X-B3-Traceid", None),
            hex_str(self.trace.trace_id),
            "There should be B3 headers if context is present",
        )



import importlib


class Local(object):
    local_ip = None


Client = None


def configure(settings, prefix):
    global Client
    transport = settings.get(prefix + "transport", "scribe")
    assert transport in ("http", "scribe")
    driver = importlib.import_module(
        "zipkin.transport.{}client".format(transport),
    )
    Client = driver.Client  # noqa
    Client.configure(settings, prefix)


def log(trace):
    try:
        Client.log(trace)
    except Exception as err:
        log.error("Unexpected Exception while sending trace: %s", err)



import threading
from .models import TraceStack

data = None


def local():
    global data
    if not data:
        data = threading.local()

    if not getattr(data, "trace", None):
        data.trace = TraceStack()
    return data.trace



def includeme(config):
    """Include the zipkin definitions"""

    # Attach the subscriber a couple of times, this allow to start logging as
    # early as possible. Later calls on the same request will enhance the more
    # we proceed through the stack (after authentication, after router, ...)
    config.include(".pyramidhook")



import logging

from zipkin import local
from zipkin.models import Annotation, Trace
from zipkin.util import hex_str, int_or_none

from zipkin.client import log


endpoint = None
logger = logging.getLogger(__name__)


def task_send_handler(body, exchange, routing_key, headers, **kwargs):
    trace = local().current

    if not trace:
        logger.warn("No zipkin parent trace found, ignoring tracing task.")
        return

    forwarded_trace = trace.child_noref("subservice")
    headers["X-B3-TraceId"] = hex_str(forwarded_trace.trace_id)
    headers["X-B3-SpanId"] = hex_str(forwarded_trace.span_id)
    forwarded_trace.record(Annotation.client_send())

    if forwarded_trace.parent_span_id is not None:
        headers["X-B3-ParentSpanId"] = hex_str(forwarded_trace.parent_span_id)


def task_prerun_handler(task_id, task, **kwargs):
    request = task.request

    trace = Trace(
        "Task execute %r" % task.name,
        int_or_none(request.headers.get("X-B3-TraceId", None)),
        int_or_none(request.headers.get("X-B3-SpanId", None)),
        int_or_none(request.headers.get("X-B3-ParentSpanId", None)),
        endpoint=endpoint,
    )

    setattr(request, "trace", trace)
    local().append(trace)
    trace.record(Annotation.server_recv())


def task_postrun_handler(task_id, task, **kwargs):
    trace = local().current
    trace.record(Annotation.server_send())
    trace.record(Annotation.client_recv())

    log(trace)
    local().pop()



from __future__ import absolute_import
import time
import logging

from pyramid.tweens import INGRESS
from pyramid.settings import aslist

from zipkin import local
from zipkin.api import stack_trace
from zipkin.models import Trace, Annotation
from zipkin.util import int_or_none
from zipkin.client import log as zipkin_log
from zipkin.config import configure as configure_zk


log = logging.getLogger(__name__)


class AllTraceTweenView(object):

    endpoint = None

    @classmethod
    def configure(cls, settings):
        default_name = "Registry"  # Keep compat with `registry.__name__` ?
        name = settings.get("zipkin.service_name", default_name)

        bindings = aslist(settings.get("zipkin.bindings", "requests celery xmlrpclib"))
        cls.endpoint = configure_zk(
            name,
            settings,
            use_requests="requests" in bindings,
            use_celery="celery" in bindings,
            use_xmlrpclib="xmlrpclib" in bindings,
        )

    def __init__(self, handler, registry):
        self.handler = handler
        self.trace = None

    def track_start_request(self, request):
        headers = request.headers

        trace_name = request.path_qs
        if request.matched_route:
            # we only get a matched route if we've gone through the router.
            trace_name = request.matched_route.pattern

        trace = Trace(
            request.method + " " + trace_name,
            int_or_none(headers.get("X-B3-TraceId", None)),
            int_or_none(headers.get("X-B3-SpanId", None)),
            int_or_none(headers.get("X-B3-ParentSpanId", None)),
            endpoint=self.endpoint,
        )

        if "X-B3-TraceId" not in headers:
            log.info("no trace info from request: %s", request.path_qs)

        if request.matchdict:  # matchdict maybe none if no route is registered
            for k, v in request.matchdict.items():
                trace.record(Annotation.string("route.param.%s" % k, v))

        trace.record(Annotation.string("http.path", request.path_qs))
        log.info("new trace %r", trace.trace_id)

        stack_trace(trace)

        trace.record(Annotation.server_recv())
        self.trace = trace

    def track_end_request(self, request, response):
        if self.trace:
            self.trace.record(Annotation.server_send())
            log.info("reporting trace %s", self.trace.name)

            response.headers["Trace-Id"] = str(self.trace.trace_id)
            zipkin_log(self.trace)

    def __call__(self, request):
        self.track_start_request(request)
        response = None
        try:
            response = self.handler(request)
        finally:
            # request.response in case an exception is raised ?
            self.track_end_request(request, response or request.response)
            local().reset()
            self.trace = None
        return response or request.response


class SlowQueryTweenView(AllTraceTweenView):

    max_duration = None

    @classmethod
    def configure(cls, settings):
        super(SlowQueryTweenView, cls).configure(settings)
        setting = settings.get("zipkin.slow_log_duration_exceed")
        if setting is None:
            log.error(
                "Missing setting 'zipkin.slow_log_duration_exceed' %r",
                list(settings.keys()),
            )
            return
        try:
            cls.max_duration = float(setting)
        except ValueError:
            log.error("Invalid setting 'zipkin.slow_log_duration_exceed'")

    def __init__(self, handler, registry):
        super(SlowQueryTweenView, self).__init__(handler, registry)
        self.start = None

    def track_start_request(self, request):
        self.start = time.time()
        super(SlowQueryTweenView, self).track_start_request(request)

    def track_end_request(self, request, response):
        if self.max_duration is None:
            # unconfigure, we don't care
            return
        if self.start:
            duration = time.time() - self.start
            if duration > self.max_duration:
                super(SlowQueryTweenView, self).track_end_request(request, response)


def includeme(config):
    """Include the zipkin definitions"""

    # Attach the subscriber a couple of times, this allow to start logging as
    # early as possible. Later calls on the same request will enhance the more
    # we proceed through the stack (after authentication, after router, ...)

    settings = config.registry.settings
    tween_factory = settings.get("zipkin.tween_factory", "all")
    assert tween_factory in ["all", "slow_query"]
    if tween_factory == "all":
        tween_factory = AllTraceTweenView
    elif tween_factory == "slow_query":
        tween_factory = SlowQueryTweenView
    else:
        log.error(
            "Invalid value for settings 'zipkin.tween_factory', should be all or slow_query, not %s",
            tween_factory,
        )
        return

    tween_factory.configure(settings)

    config.add_tween(
        "{}.{}".format(tween_factory.__module__, tween_factory.__name__),
        under=INGRESS,
    )



import logging

from zipkin import get_current_trace
from zipkin.models import Annotation

log = logging.getLogger(__name__)
endpoints = {}


def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    try:
        endpoint = endpoints.get(conn.engine)
        parent_trace = get_current_trace()

        if not parent_trace:
            log.warning("No parent found while tracing SQL")
            return

        try:
            context.trace = parent_trace.child("SQL", endpoint=endpoint)
            abstract = context
        except AttributeError:
            cursor.trace = parent_trace.child("SQL", endpoint=endpoint)
            abstract = cursor

        abstract.trace.record(Annotation.string("db.statement", statement))

        if parameters:
            if isinstance(parameters, dict):
                parameters = dict(
                    [
                        (key, getattr(param, "logged_value", param))
                        for key, param in parameters.items()
                    ]
                )
            else:
                parameters = [
                    getattr(param, "logged_value", param) for param in parameters
                ]

        abstract.trace.record(Annotation.string("db.parameters", repr(parameters)))
        abstract.trace.record(Annotation.string("span.kind", "client"))
        abstract.trace.record(Annotation.server_recv())
    except Exception:
        log.exception("Unexpected exception while tracing SQL")


def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    if not hasattr(context, "trace") and not hasattr(cursor, "trace"):
        return
    abstract = context if hasattr(context, "trace") else cursor

    try:
        abstract.trace.record(Annotation.string("status", "OK"))
        abstract.trace.record(Annotation.server_send())
    except Exception:
        log.exception("Unexpected exception while tracing SQL")


def dbapi_error(conn, cursor, statement, parameters, context, exception):
    if not hasattr(context, "trace") and not hasattr(cursor, "trace"):
        return
    abstract = context if hasattr(context, "trace") else cursor

    try:
        abstract.trace.record(Annotation.string("status", "KO"))
        abstract.trace.record(Annotation.server_send())
    except Exception:
        log.exception("Unexpected exception while tracing SQL")



from __future__ import absolute_import

try:
    import unittest2 as unittest

except ImportError:
    import unittest

import multiprocessing as mp

from six.moves.xmlrpc_server import SimpleXMLRPCServer
from six.moves.xmlrpc_client import ServerProxy

from time import time, sleep

import zipkin
from zipkin.thread import local
from zipkin.binding.xmlrpclib import bind as bind_zipkin

from ..utils import DummyClient, dummy_log

SERVERPORT = 18747


class XMLRPClibAppTestCase(unittest.TestCase):
    server = None

    def setUp(self):
        endpoint = zipkin.configure(
            "My test xmlrpclib application",
            {"zipkin.collector": "localhost", "zipkin.service_name": "my app"},
        )

        bind_zipkin(endpoint)

        server = SimpleXMLRPCServer(("localhost", SERVERPORT), allow_none=True)
        server.register_function(sleep)

        self.server = mp.Process(target=server.serve_forever)
        self.server.start()

        DummyClient.reset()

    def tearDown(self):
        if self.server and self.server.is_alive():
            self.server.terminate()

    def test_sleep(self):
        client = ServerProxy("http://localhost:%d/RPC2" % SERVERPORT, allow_none=True)

        start = time()
        client.sleep(0.1)
        end = time()
        delta = end - start

        dummy_log(local().current)

        self.assertGreater(delta, 0.101)

        traces = DummyClient._client.messages
        self.assertEqual(
            len(traces),
            1,
            "There should be one trace for " "the request just processed",
        )



from zipkin import local
from zipkin.client import log
from zipkin.models import Annotation, Trace
from zipkin.util import int_or_none

from flask import request

endpoint = None


def pre_request(app, **extra):
    headers = request.headers
    trace = Trace(
        request.method + " " + request.url,
        int_or_none(headers.get("X-B3-TraceId", None)),
        int_or_none(headers.get("X-B3-SpanId", None)),
        int_or_none(headers.get("X-B3-ParentSpanId", None)),
        endpoint=endpoint,
    )

    setattr(request, "trace", trace)
    local().append(trace)
    trace.record(Annotation.string("http.uri", request.url))
    trace.record(Annotation.server_recv())


def pre_response(app, response, **extra):
    request.trace.record(
        Annotation.string("http.responsecode", "{0}".format(response.status_code))
    )
    request.trace.record(Annotation.server_send())
    log(request.trace)
    local().pop()



import logging

from six.moves.urllib.parse import urlparse

from zipkin import local
from zipkin.models import Annotation
from zipkin.util import hex_str


log = logging.getLogger(__name__)


def filter_url_path(url):
    url = urlparse(url)._replace(path="", query=None, fragment="")
    url = url._replace(netloc=url.netloc.split("@")[-1])
    return url.geturl()


def pre_request(request):
    parent_trace = local().current
    if not parent_trace:
        return request

    url = filter_url_path(request.url)
    request.trace = parent_trace.child("requests:%s %s" % (request.method, url))
    forwarded_trace = request.trace.child_noref("subservice")

    request.headers["X-B3-TraceId"] = hex_str(forwarded_trace.trace_id)
    request.headers["X-B3-SpanId"] = hex_str(forwarded_trace.span_id)
    if forwarded_trace.parent_span_id is not None:
        request.headers["X-B3-ParentSpanId"] = hex_str(forwarded_trace.parent_span_id)

    request.trace.record(Annotation.string("http.method", request.method))
    request.trace.record(Annotation.string("http.url", request.url))
    request.trace.record(Annotation.string("span.kind", "client"))
    request.trace.record(Annotation.server_recv())

    return request


def pre_response(resp, req=None):
    if not req:
        req = resp.request

    if not hasattr(req, "trace"):
        return resp

    req.trace.record(
        Annotation.string(
            "http.status_code", "{0}".format(getattr(resp, "status", None))
        )
    )
    req.trace.record(Annotation.server_send())

    return resp


class Proxy(object):
    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    #
    # proxying (special cases)
    #
    def __getattr__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    #
    # factories
    #
    _special_names = [
        "__abs__",
        "__add__",
        "__and__",
        "__call__",
        "__cmp__",
        "__coerce__",
        "__contains__",
        "__delitem__",
        "__delslice__",
        "__div__",
        "__divmod__",
        "__eq__",
        "__float__",
        "__floordiv__",
        "__ge__",
        "__getitem__",
        "__getslice__",
        "__gt__",
        "__hash__",
        "__hex__",
        "__iadd__",
        "__iand__",
        "__idiv__",
        "__idivmod__",
        "__ifloordiv__",
        "__ilshift__",
        "__imod__",
        "__imul__",
        "__int__",
        "__invert__",
        "__ior__",
        "__ipow__",
        "__irshift__",
        "__isub__",
        "__iter__",
        "__itruediv__",
        "__ixor__",
        "__le__",
        "__len__",
        "__long__",
        "__lshift__",
        "__lt__",
        "__mod__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__oct__",
        "__or__",
        "__pos__",
        "__pow__",
        "__radd__",
        "__rand__",
        "__rdiv__",
        "__rdivmod__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__reversed__",
        "__rfloorfiv__",
        "__rlshift__",
        "__rmod__",
        "__rmul__",
        "__ror__",
        "__rpow__",
        "__rrshift__",
        "__rshift__",
        "__rsub__",
        "__rtruediv__",
        "__rxor__",
        "__setitem__",
        "__setslice__",
        "__sub__",
        "__truediv__",
        "__xor__",
        "next",
    ]

    @classmethod
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)

    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins


try:
    from requests.adapters import HTTPAdapter

    class ZipkinAdapterProxy(Proxy, HTTPAdapter):
        """This class will proxy all methods or attributes to underlying
        HTTPAdapter, it also override the send and build_response to hook
        zipkin headers and tracing.
        The proxy allows to hook into an existing HTTPAdapter
        """

        def send(self, request, *args, **kwargs):
            pre_request(request)
            return super(ZipkinAdapterProxy, self).send(request, *args, **kwargs)

        def build_response(self, req, resp, *args, **kwargs):
            pre_response(resp, req)

            return super(ZipkinAdapterProxy, self).build_response(
                req, resp, *args, **kwargs
            )

    def request_adapter(adapter):
        return ZipkinAdapterProxy(adapter)


except ImportError:
    # requests < 1.0.0
    def request_adapter(adapter):
        return adapter


def session_init(init):
    def func(self, *args, **kwargs):
        init(self, *args, **kwargs)
        if hasattr(self, "mount"):
            adapter = ZipkinAdapterProxy(HTTPAdapter())
            self.mount("http://", adapter)
            self.mount("https://", adapter)
        else:
            self.hooks["pre_request"] = pre_request
            self.hooks["response"] = pre_response

    return func



import math
import six
import time
import socket
from threading import Lock

from .util import uniq_id
from .client import Local
from .zipkin import zipkincore_thrift as constants


if not six.PY2:
    long = int


class Id(long):
    def __repr__(self):
        return "<Id %x>" % self

    def __str__(self):
        return "%x" % self


class Endpoint(object):
    """
    :param ip: C{str} ip address
    :param port: C{int} port number
    :param service_name: C{str} service_name
    """

    def __init__(self, service_name, ip=None, port=0):
        try:
            if not ip:
                if Local.local_ip:
                    ip = Local.local_ip
                else:
                    ip = socket.gethostbyname_ex(socket.gethostname())[2][0]
        except socket.gaierror:
            ip = "127.0.0.1"
        self.ip = ip
        self.port = port
        self.service_name = service_name


# TODO
# __eq__, __ne__, __repr__


class TraceStack(object):
    def __init__(self):
        self.stack = []
        self.cur = None

        # Locking is required, as stack and cur should mutate at the same time
        self.lock = Lock()

    def child(self, name, endpoint=None):
        assert isinstance(name, six.string_types), "name parameter should be a string"
        assert (
            isinstance(endpoint, Endpoint) or endpoint is None
        ), "endpoint parameter should be an Endpoint"

        try:
            trace = self.cur.child(name, endpoint)
            self.lock.acquire()
            self.stack.append(trace)
            self.cur = trace
            return trace
        finally:
            self.lock.release()

    def reset(self):
        try:
            self.lock.acquire()
            self.stack = []
            self.cur = None
        finally:
            self.lock.release()

    def replace(self, trace):
        assert isinstance(trace, Trace), "trace parameter should be of type Trace"

        try:
            self.lock.acquire()
            self.stack = [trace]
            self.cur = trace
        finally:
            self.lock.release()

    def append(self, trace):
        assert isinstance(trace, Trace), "trace parameter should be of type Trace"

        try:
            self.lock.acquire()
            self.stack.append(trace)
            self.cur = trace
        finally:
            self.lock.release()

    def pop(self):
        try:
            self.lock.acquire()

            if self.cur is None:
                raise IndexError("pop from an empty stack")

            # pop is safe here, cur is not none, current stack can't be empty
            trace = self.stack.pop()
            try:
                cur = self.stack.pop()
                self.stack.append(cur)
                self.cur = cur
            except:
                self.cur = None
            return trace
        finally:
            self.lock.release()

    @property
    def current(self):
        return self.cur


class Trace(object):
    def __init__(
        self, name, trace_id=None, span_id=None, parent_span_id=None, endpoint=None
    ):
        assert isinstance(name, six.string_types), "name parameter should be a string"
        self.name = name
        self.trace_id = Id(trace_id or uniq_id())
        self.span_id = Id(span_id or uniq_id())

        self.parent_span_id = parent_span_id

        self.annotations = []
        self._children = []

        self._endpoint = endpoint

    def record(self, *annotations):
        for a in annotations:
            if a.endpoint is None:
                a.endpoint = self._endpoint
        self.annotations.extend(annotations)

    def child_noref(self, name, endpoint=None):
        if endpoint is not None:
            e = endpoint
        else:
            e = self._endpoint
        trace = self.__class__(
            name, trace_id=self.trace_id, parent_span_id=self.span_id, endpoint=e
        )
        return trace

    def child(self, name, endpoint=None):
        trace = self.child_noref(name, endpoint)
        self._children.append(trace)
        return trace

    def children(self):
        return [y for x in self._children for y in x.children()] + [self]

    def __repr__(self):
        return "<Trace %s>" % self.trace_id


class Annotation(object):
    """
    :param name: C{str} name of this annotation.

    :param value: A value of the appropriate type based on
        C{annotation_type}.

    :param annotation_type: C{str} the expected type of our C{value}.

    :param endpoint: An optional L{IEndpoint} provider to associate with
        this annotation or C{None}
    """

    def __init__(self, name, value, annotation_type, endpoint=None):
        self.name = name
        self.value = value
        self.annotation_type = annotation_type
        self.endpoint = endpoint

    @classmethod
    def timestamp(cls, name, timestamp=None):
        if timestamp is None:
            timestamp = math.trunc(time.time() * 1000 * 1000)

        return cls(name, timestamp, "timestamp")

    @classmethod
    def server_send(cls, timestamp=None):
        return cls.timestamp(constants.SERVER_SEND, timestamp)

    @classmethod
    def server_recv(cls, timestamp=None):
        return cls.timestamp(constants.SERVER_RECV, timestamp)

    @classmethod
    def client_send(cls, timestamp=None):
        return cls.timestamp(constants.CLIENT_SEND, timestamp)

    @classmethod
    def client_recv(cls, timestamp=None):
        return cls.timestamp(constants.CLIENT_RECV, timestamp)

    @classmethod
    def string(cls, name, value):
        return cls(name, value, "string")

    @classmethod
    def bytes(cls, name, value):
        return cls(name, value, "bytes")



import os
import re
import sys
from setuptools import setup, find_packages

name = "zipkin"
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst")) as f:
    README = f.read()
with open(os.path.join(here, "CHANGES.rst")) as f:
    CHANGES = f.read()
with open(os.path.join(here, name, "__init__.py")) as version:
    VERSION = (
        re.compile(r'.*__version__ = "(.*?)"', re.S).match(version.read()).group(1)
    )

requires = [
    "thriftpy2",
    "six",
]

tests_require = ["pytest", "mock"]

if sys.version_info[:2] < (2, 7):
    tests_require.append("unittest2")

extras_require = {
    "django": [
        "django",
    ],
    "pyramid": [
        "pyramid",
    ],
    "celery": [
        "celery >= 3.1",
    ],
    "requests": [
        "requests",
    ],
    "psycopg2": [
        "psycopg2-binary",
    ],
    "flask": [
        "flask",
        "blinker",
    ],
    "dev": [
        "blinker",
        "celery",
        "django",
        "flask",
        "psycopg2-binary",
        "pyramid",
        "requests",
        "sphinx",
        "sqlalchemy",
    ],
    "test": tests_require,
}

setup(
    name="zk",
    version=VERSION,
    description="zipkin",
    long_description=README + "\n\n" + CHANGES,
    classifiers=[
        "Programming Language :: Python",
    ],
    author="Gandi",
    author_email="feedback@gandi.net",
    url="",
    keywords="zipkin metrics trace",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    extras_require=extras_require,
    tests_require=tests_require,
    test_suite="{name}.tests".format(name=name),
)



from __future__ import absolute_import

import unittest
import base64
import binascii

from ..models import Trace
from ..util import base64_thrift_formatter, int_or_none


class TestClient(unittest.TestCase):
    def test_format_64bit_traceid(self):
        _64bit_trace_id = "b1119a5629b4cb7f"
        trace = Trace("foobar", trace_id=int_or_none(_64bit_trace_id))
        annotations = []
        data = base64_thrift_formatter(trace, annotations)
        data = base64.b64decode(data)

        self.assertIn(binascii.unhexlify(_64bit_trace_id), data)

    def test_format_overflow_traceid(self):
        overflow_trace_id = "b1119a5629b4cb7fa"
        trace = Trace("foobar", trace_id=int_or_none(overflow_trace_id))
        annotations = []
        with self.assertRaises(ValueError):
            base64_thrift_formatter(trace, annotations)



import logging


from sqlalchemy import event

from . import events
from zipkin.models import Endpoint

log = logging.getLogger(__name__)


def bind(engine, endpoint=None):

    log.info("Binding zipkin to SQLALchemy")
    if not endpoint:
        endpoint = Endpoint("SQL")

    events.endpoints[engine] = endpoint
    event.listen(engine, "before_cursor_execute", events.before_cursor_execute)
    event.listen(engine, "after_cursor_execute", events.after_cursor_execute)
    # event.listen(engine, 'dbapi_error', events.dbapi_error)
    log.info("zipkin bound to SQLALchemy")



import sys
from functools import wraps

from psycopg2.extensions import connection as _connection
from psycopg2.extensions import cursor as _cursor

from zipkin.api import get_current_trace
from zipkin.models import Annotation


class TraceConnection(_connection):
    """A connection that logs all queries to a zipkin."""

    def cursor(self, *args, **kwargs):
        kwargs.setdefault("cursor_factory", self.cursor_factory or TraceCursor)
        return super(TraceConnection, self).cursor(*args, **kwargs)


def trace_req(trace_name):
    def wrapper(fn):
        @wraps(fn)
        def wrapped(cursor, statement, vars=None):
            trace = None
            parent_trace = get_current_trace()
            if parent_trace:
                trace = parent_trace.child(trace_name)
                trace.record(Annotation.string("db.statement", statement))

                if vars:
                    if isinstance(vars, dict):
                        params = dict(
                            [
                                (key, getattr(param, "logged_value", param))
                                for key, param in vars.items()
                            ]
                        )
                    else:
                        params = [
                            getattr(param, "logged_value", param) for param in vars
                        ]

                    trace.record(Annotation.string("db.parameters", repr(params)))
                    trace.record(Annotation.string("span.kind", "client"))
                    trace.record(Annotation.server_send())

            try:
                fn(cursor, statement, vars)
            finally:
                if trace:
                    status = "OK" if sys.exc_info()[0] is None else "KO"
                    trace.record(Annotation.string("status", status))
                    trace.record(Annotation.server_recv())

        return wrapped

    return wrapper


class TraceCursor(_cursor):
    """A cursor that logs queries using its connection logging facilities."""

    @trace_req("SQL")
    def execute(self, query, vars=None):
        return super(TraceCursor, self).execute(query, vars)

    @trace_req("STORED PROCEDURE")
    def callproc(self, procname, vars=None):
        return super(TraceCursor, self).callproc(procname, vars)
